from numbers import Real
import numpy as np


def beep(frequency, duration, sample_rate=44100, hanning=True):
    """Generates a sine wave.

    Args:
        frequency (int or float): The frequency of the sine wave.
        duration (int or float): Duration of the beep in seconds.

    Kwargs:
        sample_rate (int or float): Sampling rate for the wave.

    Returns:
        A 1-dimensional numpy array.

    Example:
        Generate a sine wave at 440 Hz and duration of 0.5 seconds:

        >>> my_beep = beep(440, 0.5, 44100)
    """
    sample = np.sin(2 * np.pi * frequency * (np.arange(0, int(duration * sample_rate))) / sample_rate)
    if hanning:
        apply_hanning(sample, sample_rate)
    return sample


def apply_hanning(sample, sample_rate):
    hw_size = int(min(sample_rate // 100, len(sample) // 15))
    hw = np.hanning(2 * hw_size + 1)
    sample[:hw_size] *= hw[:hw_size]
    sample[-hw_size:] *= hw[hw_size + 1:]


def beep_sequence(click_freq=(440, 660, 880, 1220),
                  inter_click_interval=0.5,
                  dur_clicks=0.04,
                  lead_in=0.1,
                  sample_rate=44100,
                  stereo=True):
    """Generate a series of sine waves, similar to a metronome.

    Kwargs:
        click_freq (number, list, tuple, or 1d numpy array): The frequency of each beep.
        inter_click_interval (int or float): The period between the beep midpoints.
        dur_clicks (int or float): Float or int, duration of each beep in seconds.
        lead_in (float): Amount of time in seconds before the center of the first beep.
        sample_rate (int or float): Sampling rate for the wave.
        stereo (bool): Generate audio for two channels.

    Returns:
        A numpy array, 1-dimensional if mono, 2-dimensional (n by 2) if stereo.

    Example:
        To generate four beeps of different frequency and spaced by half a second,

        >>> my_train = beep_sequence([1220, 400, 410, 620], inter_click_interval=0.5)
    """
    if isinstance(click_freq, Real):
        click_freq = [click_freq]
    beeps = [beep(n, duration=dur_clicks, sample_rate=sample_rate) for n in click_freq]
    space = np.zeros(int((inter_click_interval * sample_rate) - len(beeps[0])))
    out = np.zeros((int(sample_rate * lead_in - len(beeps[0]) / 2)))
    out = np.append(out, beeps[0])
    for i in range(len(click_freq) - 1):
        out = np.append(out, space)
        out = np.append(out, beeps[i + 1])
    if stereo:
        out = np.transpose(np.vstack((out, out)))
    return out



import numpy as np
from toon.input.base_input import BaseInput
from ctypes import c_double


class FakeInput(BaseInput):
    """Device used for testing."""
    def __init__(self, **kwargs):
        super(FakeInput, self).__init__(**kwargs)
        self.data_shape = FakeInput.data_shapes(**kwargs)
        self.sampling_frequency = FakeInput.samp_freq(**kwargs)

    def __enter__(self):
        self.t1 = self.clock()
        return self

    def __exit__(self, *args):
        pass

    def read(self):
        time = self.clock()
        dat = [np.random.random(ds) for ds in self.data_shape]
        if len(self.data_shape) == 1:
            dat = dat[0]
        while self.clock() < self.t1:
            pass
        self.t1 = self.clock() + (1 / self.sampling_frequency)
        return time, dat

    @staticmethod
    def samp_freq(**kwargs):
        return kwargs.get('sampling_frequency', 100)

    @staticmethod
    def data_shapes(**kwargs):
        return kwargs.get('data_shape', [[5]])

    @staticmethod
    def data_types(**kwargs):
        return kwargs.get('data_type', [c_double])



import numpy as np


def cart2pol(x, y, units='deg', ref=(0, 0)):
    """Convert cartesian coordinates to polar coordinates.

    Args:
        x
        y
    Kwargs:
        units (string): 'deg' for degrees, 'rad' for radians.
        ref (list or tuple): The reference point (x, y).
    Returns:
        (theta, radius)
    """
    x -= ref[0]
    y -= ref[1]
    radius = np.hypot(x, y)
    theta = np.arctan2(y, x)
    if units in 'deg':
        theta *= 180.0 / np.pi

    return theta, radius


def pol2cart(theta, radius, units='deg', ref=(0, 0)):
    """Convert polar coordinates to cartesian coordinates.

    Args:
        theta
        radius
    Kwargs:
        units (string): 'deg' for degrees, 'rad' for radians.
        ref (list or tuple): The reference point (x, y).
    Returns:
        (x, y)
    """
    if units in 'deg':
        theta *= np.pi / 180.0
    x = radius * np.cos(theta) + ref[0]
    y = radius * np.sin(theta) + ref[1]
    return x, y


def cart2sph(x, y, z, units='deg', ref=(0, 0, 0)):
    """Convert cartesian coordinates to spherical coordinates.

    Args:
        x
        y
        z
    Kwargs:
        units (string): 'deg' for degrees, 'rad' for radians.
        ref (list or tuple): The reference point (x, y, z).
    Returns:
        (elevation, azimuth, radius)
    """
    x -= ref[0]
    y -= ref[1]
    z -= ref[2]
    radius = np.sqrt(x ** 2 + y ** 2 + z ** 2)
    azimuth = np.arctan2(y, x)
    elevation = np.arctan2(z, np.sqrt(x ** 2 + y ** 2))
    if units in 'deg':
        azimuth *= 180.0 / np.pi
        elevation *= 180.0 / np.pi
    return elevation, azimuth, radius


def sph2cart(elevation, azimuth, radius, units='deg', ref=(0, 0, 0)):
    """Convert spherical coordinates to cartesian coordinates.

    Args:
        elevation
        azimuth
        radius
    Kwargs:
        units (string): 'deg' for degrees, 'rad' for radians.
        ref (list or tuple): The reference point (x, y, z).
    Returns:
        (x, y, z)
    """
    if units in 'deg':
        azimuth *= np.pi / 180.0
        elevation *= np.pi / 180.0

    x = radius * np.cos(elevation) * np.cos(azimuth) + ref[0]
    y = radius * np.cos(elevation) * np.sin(azimuth) + ref[1]
    z = radius * np.sin(elevation) + ref[2]

    return x, y, z



import multiprocessing as mp
import ctypes
import gc
import os
import psutil
import numpy as np
from toon.input.helper import check_and_fix_dims, shared_to_numpy


class MultiprocessInput(object):
    def __init__(self, device=None, high_priority=True, use_gc=False,
                 nrow=None, **kwargs):
        """
        Args:
            device: A class that inherits from `toon.input.base_input.BaseInput` (not an object).
            high_priority (bool): Whether the remote process should run at high priority.
                Silently fails if permissions are insufficient.
            use_gc (bool): Optionally disable garbage collection on the remote process.
            nrow: Override the inferred number of rows to store.
                If `None`, defaults to 10x the amount expected from a single frame.
            **kwargs: Passed to the device during instantiation.
        """
        self.device = device
        self.device_args = kwargs
        self.nrow = nrow
        self.high_priority = high_priority
        self.use_gc = use_gc

    def start(self):
        # Use spawn, rather than fork, when possible (for Macs)
        try:
            mp.set_start_method('spawn')
        except (AttributeError, RuntimeError):
            pass
        self.shared_lock = mp.RLock()
        self.remote_ready = mp.Event()
        self.kill_remote = mp.Event()
        self.sample_count = mp.Value(
            ctypes.c_uint32, 0, lock=self.shared_lock)  # sample counter
        data_dims = check_and_fix_dims(
            self.device.data_shapes(**self.device_args))
        # TODO: unclear, change later
        self.not_time_axis = [
            tuple(range(-1, -(len(dd) + 1), -1)) for dd in data_dims]
        # if not manually overridden, preallocate 10*whatever we would expect in a single frame
        nrow = self.nrow
        if not nrow:
            nrow = int(np.ceil(self.device.samp_freq(
                **self.device_args) / 60.0) * 10)

        # preallocate data arrays (shared arrays, numpy equivalents, and local copies)
        data_types = self.device.data_types(**self.device_args)
        self.mp_data_arrays = []
        self.np_data_arrays = []
        self.local_data_arrays = []
        for counter, data_dim in enumerate(data_dims):
            data_dim.insert(0, nrow)  # make axis 0 time
            prod = int(np.prod(data_dim))
            data_type = data_types[counter]
            self.mp_data_arrays.append(mp.Array(data_type,
                                                prod,
                                                lock=self.shared_lock))
            self.np_data_arrays.append(shared_to_numpy(
                self.mp_data_arrays[counter], data_dim, data_type))
            self.np_data_arrays[counter].fill(np.nan)
            self.local_data_arrays.append(
                np.copy(self.np_data_arrays[counter]))

        # preallocate time array
        self.mp_time_array = mp.Array(
            ctypes.c_double, nrow, lock=self.shared_lock)
        self.np_time_array = shared_to_numpy(
            self.mp_time_array, nrow, ctypes.c_double)
        self.np_time_array.fill(np.nan)
        self.local_time_array = np.copy(self.np_time_array)

        #
        self.process = mp.Process(target=worker,
                                  args=(self.device,
                                        self.device_args,
                                        self.shared_lock,
                                        self.remote_ready,
                                        self.kill_remote,
                                        self.mp_time_array,
                                        self.mp_data_arrays,
                                        nrow, data_dims, data_types, self.use_gc,
                                        self.sample_count,
                                        os.getpid()))
        self.process.daemon = True
        self.process.start()
        self.ps_process = psutil.Process(self.process.pid)
        self.original_nice = self.ps_process.nice()
        self.set_high_priority(self.high_priority)
        self.remote_ready.wait()

    def read(self):
        """

        Returns:
            A (time, data) tuple, where time is a 1-D numpy array of timestamps, and
                data is either a N-D array (where the 0th axis is time), or a list of
                N-D arrays. If there is no data, returns (None, None).
        """
        with self.shared_lock:
            count = self.sample_count.value
            self.sample_count.value = 0
            np.copyto(self.local_time_array, self.np_time_array)
            for local_data, remote_data in zip(self.local_data_arrays, self.np_data_arrays):
                np.copyto(local_data, remote_data)
        if not count:
            return None, None
        dims = self.not_time_axis
        time = self.local_time_array[0:count]
        if len(dims) == 1:
            data = self.local_data_arrays[0][0:count, :]
        else:
            data = [local_data[0:count, :]
                    for local_data, dim in zip(self.local_data_arrays, dims)]
        return time, data

    def stop(self, *args):
        self.set_high_priority(False)
        self.kill_remote.set()

    def set_high_priority(self, val):
        try:
            if val:
                if psutil.WINDOWS:
                    self.ps_process.nice(psutil.HIGH_PRIORITY_CLASS)
                else:
                    self.ps_process.nice(-10)
            else:
                self.ps_process.nice(self.original_nice)
        except (psutil.AccessDenied, psutil.NoSuchProcess):
            pass


def worker(device, device_args, shared_lock, remote_ready,
           kill_remote, mp_time_array, mp_data_arrays,
           nrow, data_dims, data_types, use_gc, sample_count,
           parent_pid):
    if not use_gc:
        gc.disable()
    dev = device(**device_args)
    np_time_array = shared_to_numpy(mp_time_array, nrow, ctypes.c_double)
    np_data_arrays = []
    for data, dims, types in zip(mp_data_arrays, data_dims, data_types):
        np_data_arrays.append(shared_to_numpy(data, dims, types))
    with dev as d:
        remote_ready.set()
        while not kill_remote.is_set() and psutil.pid_exists(parent_pid):
            timestamp, data = d.read()
            if timestamp is not None:
                with shared_lock:
                    if sample_count.value < nrow:  # nans to fill in
                        next_index = sample_count.value
                        if isinstance(data, list):
                            for np_data, new_data in zip(np_data_arrays, data):
                                np_data[next_index, :] = new_data
                        else:
                            np_data_arrays[0][next_index, :] = data
                        np_time_array[next_index] = timestamp
                    else:  # replace oldest data with newest data
                        if isinstance(data, list):
                            for np_data, new_data in zip(np_data_arrays, data):
                                np_data[:] = np.roll(np_data, -1, axis=0)
                                np_data[-1, :] = new_data
                        else:
                            np_data_arrays[0][:] = np.roll(
                                np_data_arrays[0], -1, axis=0)
                            np_data_arrays[0][-1, :] = data
                        np_time_array[:] = np.roll(np_time_array, -1, axis=0)
                        np_time_array[-1] = timestamp
                    sample_count.value += 1



from ctypes import c_double

import nidaqmx
import numpy as np
from nidaqmx.constants import AcquisitionType, TerminalConfiguration
from nidaqmx.stream_readers import AnalogMultiChannelReader

from toon.input.base_input import BaseInput


class ForceKeyboard(BaseInput):
    """1-DoF force transducers."""

    @staticmethod
    def samp_freq(**kwargs):
        return kwargs.get('sampling_frequency', 200)

    @staticmethod
    def data_shapes(**kwargs):
        return [[10]]

    @staticmethod
    def data_types(**kwargs):
        return [c_double]

    def __init__(self, **kwargs):
        super(ForceKeyboard, self).__init__(**kwargs)
        self.sampling_frequency = ForceKeyboard.samp_freq(**kwargs)
        self.period = 1/self.sampling_frequency
        self.t1 = 0
        self._data_buffer = np.full(ForceKeyboard.data_shapes(**kwargs)[0], np.nan)

    def __enter__(self):
        # assume first NI DAQ is the one we want
        self._device_name = nidaqmx.system.System.local().devices[0].name
        self._channels = [self._device_name + '/ai' + str(n) for n in
                          [2, 9, 1, 8, 0, 10, 3, 11, 4, 12]]
        self._device = nidaqmx.Task()
        self._device.ai_channels.add_ai_voltage_chan(
            ','.join(self._channels),
            terminal_config=TerminalConfiguration.RSE
        )
        self._reader = AnalogMultiChannelReader(self._device.in_stream)
        self._device.start()
        return self

    def read(self):
        self._reader.read_one_sample(self._data_buffer)
        time = self.clock()
        while self.clock() < self.t1:
            pass
        self.t1 = self.clock() + self.period
        return time, self._data_buffer

    def __exit__(self, *args):
        self._device.stop()
        self._device.close()



import struct
from ctypes import c_double

import hid
import numpy as np
from toon.input.base_input import BaseInput


class Hand(BaseInput):
    """Hand Articulation Neuro-Training Device (HAND)."""

    @staticmethod
    def samp_freq(**kwargs):
        return kwargs.get('sampling_frequency', 1000)

    @staticmethod
    def data_shapes(**kwargs):
        return [[15]]

    @staticmethod
    def data_types(**kwargs):
        return [c_double]

    def __init__(self, nonblocking=False, **kwargs):
        super(Hand, self).__init__(**kwargs)
        self._sqrt2 = np.sqrt(2)
        self.nonblocking = nonblocking
        self._device = None
        self._data_buffer = np.full(15, np.nan)

    def __enter__(self):
        self._device = hid.device()
        dev_path = next((dev for dev in hid.enumerate()
                         if dev['vendor_id'] == 0x16c0 and dev['interface_number'] == 0), None)['path']
        self._device.open_path(dev_path)
        self._device.set_nonblocking(self.nonblocking)
        return self

    def __exit__(self, *args):
        self._device.close()

    def read(self):
        data = self._device.read(46)
        time = self.clock()
        if data:
            # Timestamp, deviation, and 20 unsigned ints
            data = struct.unpack('>Lh' + 'H'*20, bytearray(data))
            data = np.array(data, dtype='d')
            data[0] /= 1000.0
            data[2:] /= 65535.0
            self._data_buffer[0::3] = (data[2::4] - data[3::4])/self._sqrt2  # x
            self._data_buffer[1::3] = (data[2::4] + data[3::4])/self._sqrt2  # y
            self._data_buffer[2::3] = data[4::4] + data[5::4]  # z
            return time, self._data_buffer



from numbers import Number
import numpy as np


def check_and_fix_dims(input):
    """
    Helper function to ensure data dimensions are consistent and unambiguous.
    Args:
        input: Scalar, list, or list of lists.
    Returns:
        List of lists.
    """
    # handle special-case, single scalar
    if isinstance(input, Number):
        input = [[input]]
    elif isinstance(input, (list, tuple, np.ndarray)):
        # special-case num 2, where we have a single scalar in a list
        if len(input) == 1 and isinstance(input[0], Number):
            input = [input]
        elif len(input) != 1 and any([isinstance(x, Number) for x in input]):
            raise ValueError('Ambiguous dimensions. There should be one list per expected piece of data' + \
                             ' from the input device.')
        # coerce array-like things to lists
        input = [list(x) for x in input]
        # now we're relatively comfortable we have a list of lists
    else:
        raise ValueError('Something is wrong with the input.')
    return input


def shared_to_numpy(mp_arr, dims, dtype):
    """Convert a :class:`multiprocessing.Array` to a numpy array.
    Helper function to allow use of a :class:`multiprocessing.Array` as a numpy array.
    Derived from the answer at:
    <https://stackoverflow.com/questions/7894791/use-numpy-array-in-shared-memory-for-multiprocessing>
    """
    return np.frombuffer(mp_arr.get_obj(), dtype=dtype).reshape(dims)


from ctypes import c_int32
from toon.input.base_input import BaseInput
from pynput import mouse
import numpy as np


class Mouse(BaseInput):
    """Mouse interface."""
    @staticmethod
    def samp_freq(**kwargs):
        return kwargs.get('sampling_frequency', 100)

    @staticmethod
    def data_shapes(**kwargs):
        return [[2]]

    @staticmethod
    def data_types(**kwargs):
        return [c_int32]

    def __init__(self, **kwargs):
        super(Mouse, self).__init__(**kwargs)

    def __enter__(self):
        self.dev = mouse.Listener(on_move=self.on_move)
        self.dev.start()
        self.dev.wait()
        self.times = []
        self.readings = []
        self.x_prev = 0
        self.y_prev = 0
        return self

    def __exit__(self, *args):
        self.dev.stop()
        self.dev.join()

    def on_move(self, x, y):
        self.times.append(self.clock())
        self.readings.append(np.array([x - self.x_prev, y - self.y_prev]))
        self.x_prev = x
        self.y_prev = y

    def read(self):
        """Returns the *change* in mouse position, not the absolute position."""
        if not self.readings:
            return None, None
        time2 = []
        read2 = []
        time2[:] = self.times
        read2[:] = self.readings
        self.readings = []
        return time2[-1], read2[-1]  # hack until multi-datapoint



import abc
import six
from toon.input.clock import mono_clock

@six.add_metaclass(abc.ABCMeta)
class BaseInput():
    """Abstract base class for input devices."""
    @abc.abstractmethod
    def __init__(self, clock=mono_clock.get_time, **kwargs):
        """
        Args:
            clock: A function that returns the current time. Defaults to :obj:`toon.input.clock.mono_clock.get_time`.
            **kwargs: Device-specific keyword arguments. These can also be used to infer the sampling frequency,
                shape of the data, and type of data.
        """
        self.clock = clock

    @abc.abstractmethod
    def __enter__(self):
        """Start communication with the device here."""
        pass

    @abc.abstractmethod
    def __exit__(self, *args):
        """Gracefully clean up the device here (if necessary)."""
        pass

    @abc.abstractmethod
    def read(self):
        """Read a single measurement from the input device.

        Returns:
            (timestamp, data) tuple. See examples for details.
        """
        pass

    @abc.abstractmethod
    def samp_freq(**kwargs):
        """Infer the sampling frequency from keyword arguments."""
        pass

    @abc.abstractmethod
    def data_shapes(**kwargs):
        """Infer the shape of the data from keyword arguments.

        Must be a list of lists, e.g. [[2, 3], [8]] or [[5]].
        """
        pass

    @abc.abstractmethod
    def data_types(**kwargs):
        """Infer the type of the data from keyword arguments.

        Please use types from :obj:`ctypes`.

        Must be a list, e.g. [ctypes.c_int32, ctypes.c_double] or [ctypes.c_char].
        """
        pass



from setuptools import setup, find_packages
from codecs import open
from os import path
import platform

here = path.abspath(path.dirname(__file__))

# get requirements
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = f.read().splitlines()

setup(
    name='toon',
    version='0.9.0',
    description='Tools for neuroscience experiments',
    url='https://github.com/aforren1/toon',
    author='Alexander Forrence',
    author_email='aforren1@jhu.edu',
    license='GPL3',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6'
    ],
    install_requires=requirements,
    extras_require={
        'full': ['hidapi', 'pyserial', 'pynput', 'nidaqmx;platform_system=="Windows"'],
        'hand': ['hidapi'],
        'birds': ['pyserial'],
        'keyboard': ['pynput'],
        'force': ['nidaqmx;platform_system=="Windows"']
    },
    keywords='psychophysics neuroscience input experiment',
    packages=find_packages(exclude=['contrib', 'docs', 'tests'])
)



import os

if os.name is 'nt':
    from ctypes import byref, c_int64, windll
    _fcounter = c_int64()
    _qpfreq = c_int64()
    windll.Kernel32.QueryPerformanceFrequency(byref(_qpfreq))
    _qpfreq = float(_qpfreq.value)
    _winQPC = windll.Kernel32.QueryPerformanceCounter

    def get_time():
        _winQPC(byref(_fcounter))
        return _fcounter.value / _qpfreq
else:
    from timeit import default_timer as get_time

class MonoClock(object):
    """A stripped-down version of psychopy's clock.MonotonicClock.

    I wanted to avoid importing pyglet on the remote process, in case that causes any headache.
    """
    def __init__(self, start_time=None):
        if not start_time:
            # this is sub-millisec timer in python
            self._start_time = get_time()
        else:
            self._start_time = start_time
    def get_time(self):
        """Returns the current time on this clock in secs (sub-ms precision)
        """
        return get_time() - self._start_time
    def getTime(self):
        """Alias get_time so we can set the default psychopy clock
        """
        return self.get_time()
    @property
    def start_time(self):
        return self._start_time

mono_clock = MonoClock()
    


from toon.input.base_input import BaseInput
from pynput import keyboard
from ctypes import c_bool, c_uint16, c_char


class Keyboard(BaseInput):
    """Keyboard interface."""

    @staticmethod
    def samp_freq(**kwargs):
        return kwargs.get('sampling_frequency', 100)

    @staticmethod
    def data_shapes(**kwargs):
        return [[1], [1], [1]]

    @staticmethod
    def data_types(**kwargs):
        return [c_bool, c_uint16, c_char]

    def __init__(self, keys=None, **kwargs):
        """
        Args:
            keys (list): Keys of interest, e.g. `['a', 's', 'd', 'f']`
            **kwargs: Keyword arguments passed to `toon.input.base_input.BaseInput`.
        """
        self.keys = keys
        super(Keyboard, self).__init__(**kwargs)

    def __enter__(self):
        self._on = []
        self.pressed = []
        self.index = []
        self.char = []
        self.times = []
        self.dev = keyboard.Listener(on_press=self.on_press,
                                     on_release=self.on_release)
        self.dev.start()
        self.dev.wait()
        return self

    def __exit__(self, *args):
        self.dev.stop()
        self.dev.join()

    def read(self):
        """Returns all key press/release events since the last `read()` call.

        The first data element is boolean (whether key event was press or release),
        the second element is the position of the key in the `keys` list provided at instantiation,
        and the third element is the character representation of the key."""
        if not self.times:
            return None, None
        time2 = []
        pressed2 = []
        index2 = []
        char2 = []
        time2[:] = self.times
        pressed2[:] = self.pressed
        index2[:] = self.index
        char2[:] = self.char
        self.times = []
        self.pressed = []
        self.index = []
        self.char = []
        return time2[-1], [pressed2[-1], index2[-1], char2[-1]]

    def on_press(self, key):
        ts = self.clock()
        if not isinstance(key, keyboard.Key):
            if key.char in self.keys and key.char not in self._on:
                index = self.keys.index(key.char)
                self.pressed.append(True)
                self.index.append(index)
                self.char.append(key.char)
                self.times.append(ts)
                self._on.append(key.char)

    def on_release(self, key):
        ts = self.clock()
        if not isinstance(key, keyboard.Key):
            if key.char in self.keys and key.char in self._on:
                index = self.keys.index(key.char)
                self.pressed.append(False)
                self.index.append(index)
                self.char.append(key.char)
                self.times.append(ts)
                self._on.remove(key.char)
